#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <assert.h>
#include <stdint.h>
#include <sys/mman.h>
#include <sys/ioctl.h>		/* ioctl */
#include "vcio.h"

#define PAGE_SIZE (4*1024)

int debug=0,cache=0,offset=0;

char image[1024]="a.out.text";

static void *mapmem(unsigned base, unsigned size)
{
   int mem_fd;
   unsigned offset = base % PAGE_SIZE; 
   base = base - offset;
   /* open /dev/mem */
   if ((mem_fd = open("/dev/mem", O_RDWR|O_SYNC) ) < 0) {
      printf("can't open /dev/mem \n");
      exit (-1);
   }
   void *mem = mmap(
      0,
      size,
      PROT_READ|PROT_WRITE,
      MAP_SHARED/*|MAP_FIXED*/,
      mem_fd, 
      base);
#ifdef DEBUG
   printf("base=0x%x, mem=%p\n", base, mem);
#endif
   if (mem == MAP_FAILED) {
      printf("mmap error %d\n", (int)mem);
      exit (-1);
   }
   close(mem_fd);
   return (char *)mem + offset;
}

static void *unmapmem(void *addr, unsigned size)
{
   int s = munmap(addr, size);
   if (s != 0) {
      printf("munmap error %d\n", s);
      exit (-1);
   }
}

/* 
 * use ioctl to send mbox property message
 */

static int mbox_property(int file_desc, void *buf)
{
   int ret_val = ioctl(file_desc, IOCTL_MBOX_PROPERTY, buf);

   if (ret_val < 0) {
      printf("ioctl_set_msg failed:%d\n", ret_val);
   }

#ifdef DEBUG
   unsigned *p = message; int i; unsigned size = *(unsigned *)buf;
   for (i=0; i<size/4; i++)
      printf("%04x: 0x%08x\n", i*sizeof *p, p[i]);
#endif
   return ret_val;
}

static unsigned get_version(int file_desc)
{
   int i=0;
   unsigned p[32];
   p[i++] = 0; // size
   p[i++] = 0x00000000; // process request

   p[i++] = 0x00000001; // get firmware version
   p[i++] = 0x00000004; // buffer size
   p[i++] = 0x00000000; // request size
   p[i++] = 0x00000000; // value buffer

   p[i++] = 0x00000000; // end tag
   p[0] = i*sizeof *p; // actual size

   mbox_property(file_desc, p);
   return p[5];
}

unsigned mem_alloc(int file_desc, unsigned size, unsigned align, unsigned flags)
{
   int i=0;
   unsigned p[32];
   p[i++] = 0; // size
   p[i++] = 0x00000000; // process request

   p[i++] = 0x3000c; // (the tag id)
   p[i++] = 12; // (size of the buffer)
   p[i++] = 12; // (size of the data)
   p[i++] = size; // (num bytes? or pages?)
   p[i++] = align; // (alignment)
   p[i++] = flags; // (MEM_FLAG_L1_NONALLOCATING)

   p[i++] = 0x00000000; // end tag
   p[0] = i*sizeof *p; // actual size

   mbox_property(file_desc, p);
   return p[5];
}

unsigned mem_free(int file_desc, unsigned handle)
{
   int i=0;
   unsigned p[32];
   p[i++] = 0; // size
   p[i++] = 0x00000000; // process request

   p[i++] = 0x3000f; // (the tag id)
   p[i++] = 4; // (size of the buffer)
   p[i++] = 4; // (size of the data)
   p[i++] = handle;

   p[i++] = 0x00000000; // end tag
   p[0] = i*sizeof *p; // actual size

   mbox_property(file_desc, p);
   return p[5];
}

unsigned mem_lock(int file_desc, unsigned handle)
{
   int i=0;
   unsigned p[32];
   p[i++] = 0; // size
   p[i++] = 0x00000000; // process request

   p[i++] = 0x3000d; // (the tag id)
   p[i++] = 4; // (size of the buffer)
   p[i++] = 4; // (size of the data)
   p[i++] = handle;

   p[i++] = 0x00000000; // end tag
   p[0] = i*sizeof *p; // actual size

   mbox_property(file_desc, p);
   return p[5];
}

unsigned mem_unlock(int file_desc, unsigned handle)
{
   int i=0;
   unsigned p[32];
   p[i++] = 0; // size
   p[i++] = 0x00000000; // process request

   p[i++] = 0x3000e; // (the tag id)
   p[i++] = 4; // (size of the buffer)
   p[i++] = 4; // (size of the data)
   p[i++] = handle;

   p[i++] = 0x00000000; // end tag
   p[0] = i*sizeof *p; // actual size

   mbox_property(file_desc, p);
   return p[5];
}
void printbuf(unsigned *p)
{
 int i;
 for(i=0;i<32;i++) printf("%d:%x\n",i,p[i]);
}


unsigned execute_code(int file_desc, unsigned code, unsigned r0, unsigned r1, unsigned r2, unsigned r3, unsigned r4, unsigned r5)
{
   int i=0;
   unsigned p[32];
   p[i++] = 0; // size
   p[i++] = 0x00000000; // process request

   p[i++] = 0x30010; // (the tag id)
   p[i++] = 28; // (size of the buffer)
   p[i++] = 28; // (size of the data)
   p[i++] = code;
   p[i++] = r0;
   p[i++] = r1;
   p[i++] = r2;
   p[i++] = r3;
   p[i++] = r4;
   p[i++] = r5;

   p[i++] = 0x00000000; // end tag
   p[0] = i*sizeof *p; // actual size

//printbuf(p);

   mbox_property(file_desc, p);

//printbuf(p);

   return p[5];
}

// test code for peeking and poking with mmap
int xmain(int argc, char **argv)
{
   int i;
   unsigned base;
   if (argc > 2) {
      base = strtoul(argv[1], NULL, 0);
      volatile unsigned *p = mapmem(base, PAGE_SIZE);
      p[0] = strtoul(argv[2], NULL, 0);
      for (i=0; i<1; i++)
         printf("%04x: 0x%08x\n", i*sizeof *p, p[i]);
      return 0;
   }
   if (argc > 1) {
      base = strtoul(argv[1], NULL, 0);
      volatile unsigned *p = mapmem(base, PAGE_SIZE);
      for (i=0; i<8; i++)
         printf("%04x: 0x%08x\n", i*sizeof *p, p[i]);
      return 0;
   }
   return 0;
}

void db_print(void *user,unsigned buffer,int p,int size,int cnt)
{
  int i;
  if(size==4) size&=~3;
  if(size==2) size&=~1;
  for(i=0;i<cnt;i++){
    if(((i*size)&15)==0)
      printf("\n%08x:",p);
    if(size==1)
      printf(" %02x",*(unsigned char*)(((char*)user)+p));
    else if(size==2)
      printf(" %04x",*(unsigned short*)(((char*)user)+p));
    else
      printf(" %08x",*(int*)(((char*)user)+p));
    p+=size;
  }
  if(size==4&&cnt==1)
    printf(" [%08x]",(*(int*)(((char*)user)+p-4))-buffer);
  printf("\n");
}

void db_break(void *user,unsigned buffer,int p)
{
  int offset;
  unsigned short *m;
  offset=(0x4e - p)/2;
  m=user;
  m+=p/2;
  m[0]=0x9000|(((offset>>23)&0xf)<<8)|((offset>>16)&0x7f)|0x80;
  m[1]=offset;
}

/* 
 * Main - Call the ioctl functions 
 */
int main(int argc, char *argv[])
{
  FILE *fp;
  int file_desc, i, resume = 0;
  unsigned int a0,a1,a2,a3,a4,a5;
  unsigned handle, buffer, size;
  void *user;

  for(i=1;i<argc;i++){
    if(argv[i][0]=='-'){
      if(!strcmp(argv[i],"-debug"))
        debug=1;
      else if(!strcmp(argv[i],"-cache"))
        cache=1;
      else if(!strcmp(argv[i],"-offset"))
        offset=1024;
      else
        fprintf(stderr,"ignored unknown option <%s>!\n",argv[i]);
    }else
      strcpy(image,argv[i]);
  }
    

  // open a char device file used for communicating with kernel mbox driver
  file_desc = open(DEVICE_FILE_NAME, 0);
  if (file_desc < 0) {
    printf("Can't open device file: %s\n", DEVICE_FILE_NAME);
    printf("Try creating a device file with: mknod %s c %d 0\n", DEVICE_FILE_NAME, MAJOR_NUM);
    exit(-1);
  }
  // sanity check we care talking to GPU - should match vcgencmd version
  if(debug) printf("Version: %d\n", get_version(file_desc));


  // open image, get size, allocate memory, load image
  fp = fopen(image, "rb");

  if(!fp){
    fprintf(stderr,"Could not open <%s>!\n","a.out.text");
    exit(EXIT_FAILURE);
  }

  if(fseek(fp,0,SEEK_END)){
    fprintf(stderr,"Could not determine file size!\n");
    exit(EXIT_FAILURE);
  }

  size=ftell(fp);

  if(debug) printf("memory size: %08x\n",size);
  handle = mem_alloc(file_desc, size+offset, 4096, 0xc);
  buffer = mem_lock(file_desc, handle);
  user = mapmem(buffer, size+offset);
  if(debug) printf("handle=%x, buffer=%x user=%x\n", handle, buffer,user);
  memset(user, 0, size+offset);

  fseek(fp,0,SEEK_SET);
  if(fread(user+offset, 1, size, fp)!=size){
    fprintf(stderr,"warning: image not completely read\n");
  }
  fclose(fp);

  // open reloc file, do relocations
  strcat(image,".reltext");
  fp=fopen(image,"rb");
  if(fp){
    fread(&a1,1,4,fp);
    if(debug) printf("relocs=%d\n",a1);
    for(i=0;i<a1;i++){
      int *p;
      fread(&a2,1,4,fp);
      if(debug) printf("reloc[%d]=%08x\n",i,a2);
      p=(int*)(((char*)(user))+a2+offset);
      if(debug) printf("orig=%08x\n",*p);
      (*p)+=buffer+offset;
    }
    fclose(fp);
  }else{
    fprintf(stderr,"warning: no relocs found\n");
  }

  a0=buffer;

  do{

    int p;

    while(debug){
      char cmd[1024];
      int p,cnt;
      printf("\n> ");fflush(stdout);
      gets(cmd);
      if(!strcmp(cmd,"c")) break;
      if(!strcmp(cmd,"q")){
        resume=2;
        break;
      }
      if(sscanf(cmd,"w %i %i",&p,&cnt)==2){
	db_print(user,buffer,p,4,cnt);
	continue;
      }
      if(sscanf(cmd,"w %i",&p)==1){
	db_print(user,buffer,p,4,1);
	continue;
      }
      if(sscanf(cmd,"h %i %i",&p,&cnt)==2){
	db_print(user,buffer,p,2,cnt);
	continue;
      }
      if(sscanf(cmd,"h %i",&p)==1){
	db_print(user,buffer,p,2,1);
	continue;
      }
      if(sscanf(cmd,"b %i %i",&p,&cnt)==2){
	db_print(user,buffer,p,1,cnt);
	continue;
      }
      if(sscanf(cmd,"b %i",&p)==1){
	db_print(user,buffer,p,1,1);
	continue;
      }
      if(sscanf(cmd,"bp %i",&p)==1){
	db_break(user,buffer,p);
	continue;
      }
      printf("illegal command!\n");
    }

    if(resume==2) break;

    // execute the GPU code:
    int s = execute_code(file_desc, buffer+(resume?2:0)+cache+offset, a0, a1, a2, a3, a4, a5);

    resume=1;

#define OSC (offset/4)

    if(((int *)user)[1+OSC]==4){
      if(debug) printf("write(%d,%p,%d)\n",((int *)user)[2+OSC],((int *)user)[3+OSC],((int *)user)[4+OSC]);
      a0 = write(((int *)user)[2+OSC],((int *)user)[3+OSC]-buffer+user,((int *)user)[4+OSC]);
    }else if(((int *)user)[1+OSC]==3){
      if(debug) printf("read(%d,%p,%d)\n",((int *)user)[2+OSC],((int *)user)[3+OSC],((int *)user)[4+OSC]);
      a0 = read(((int *)user)[2+OSC],((int *)user)[3+OSC]-buffer+user,((int *)user)[4+OSC]);
    }else if(((int *)user)[1+OSC]==13){
      if(debug) printf("time(%p)\n",((int*)user)[2+OSC]);
      if(!((int*)user)[2+OSC])
        a0 = time(0);
      else
        a0 = time(((int *)user)[3+OSC]-buffer+user);
    }else if(((int *)user)[1+OSC]==1){
      if(debug) printf("exit(%p)\n",((int*)user)[2+OSC]);
    }else{
      printf("unsupported syscall %d(%08x,%08x,%08x,%08x,%08x)\n",((int *)user)[1+OSC],((int *)user)[2+OSC],((int *)user)[3+OSC],((int *)user)[4+OSC],((int *)user)[5+OSC],((int *)user)[6+OSC]);
    }

  }while(((int *)user)[1+OSC]!=1&&resume!=2);

  munmap(user, size+offset);
  mem_unlock(file_desc, handle);
  mem_free(file_desc, handle);

  close(file_desc);
  return 0;
}
